from aiosnow.query.condition import Condition
from aiosnow.query.operators import IntegerOperator

from .base import BaseQueryable


class IntegerQueryable(BaseQueryable):
    def equals(self, value: int) -> Condition:
        """
        Example: reassignment_count.equals(0)

        All records in which the Reassignment count is nothing else but 0.
        """

        return self._condition(
            IntegerOperator.EQUALS, value, field_operator=IntegerOperator.SAME
        )

    def not_equals(self, value: int) -> Condition:
        """
        Example: reassignment_count.not_equals(0)

        All records in which the value for the Reassignment count is any number but 0.
        """

        return self._condition(
            IntegerOperator.NOT_EQUALS, value, field_operator=IntegerOperator.DIFFERENT
        )

    def less_than(self, value: int) -> Condition:
        """
        Example: impact.less_than(2)

        All records in which the Impact field has a value of 1 - High.
        """

        return self._condition(
            IntegerOperator.LESS, value, field_operator=IntegerOperator.Related.LESS
        )

    def greater_than(self, value: int) -> Condition:
        """
        Example: impact.greater_than(2)

        All records in which the Impact field has a value of 3 - Low
        """

        return self._condition(
            IntegerOperator.GREATER,
            value,
            field_operator=IntegerOperator.Related.GREATER,
        )

    def less_or_equals(self, value: int) -> Condition:
        """
        Example: impact.less_or_equals(2)

        All records in which the Impact field has a value of 1 - High or 2 - Medium.
        """

        return self._condition(
            IntegerOperator.LESS_EQUALS,
            value,
            field_operator=IntegerOperator.Related.LESS_EQUALS,
        )

    def greater_or_equals(self, value: int) -> Condition:
        """
        Example: impact.greater_or_equals(2)

        All records in which the Impact field has a value of 2 - Medium or 3 - Low.
        """

        return self._condition(
            IntegerOperator.GREATER_EQUALS,
            value,
            field_operator=IntegerOperator.Related.GREATER_EQUALS,
        )

    def between(self, value1: int, value2: int) -> Condition:
        """
        Example: impact.between([1, 3])

        All records in which the Impact field has one of the following values:
        1 - High
        2 - Medium
        3 - Low
        """

        value = f"{value1}@{value2}"
        return self._condition(IntegerOperator.BETWEEN, value)



class BaseOperator:
    EQUALS = "="
    NOT_EQUALS = "!="
    LESS = "<"
    LESS_EQUALS = "<="
    GREATER = ">"
    GREATER_EQUALS = ">="
    CONTAINS = "*"
    NOT_CONTAINS = "!*"
    EMPTY = "ISEMPTY"
    POPULATED = "ISNOTEMPTY"
    BETWEEN = "BETWEEN"
    ANYTHING = "ANYTHING"
    SAME = "SAMEAS"
    DIFFERENT = "NSAMEAS"
    ONEOF = "IN"
    NOT_ONEOF = "NOT IN"


class StringOperator(BaseOperator):
    EMPTYSTRING = "EMPTYSTRING"
    STARTSWITH = "STARTSWITH"
    ENDSWITH = "ENDSWITH"


class ReferenceField(StringOperator):
    pass


class IntegerOperator(BaseOperator):
    class Related:
        GREATER = "GT_FIELD"
        LESS = "LT_FIELD"
        GREATER_EQUALS = "GT_OR_EQUALS_FIELD"
        LESS_EQUALS = "LE_OR_EQUALS_FIELD"


class DateTimeOperator(BaseOperator):
    ON = "ON"
    NOT_ON = "NOTON"
    TREND = "DATEPART"
    RELATIVE_GREATER = "RELATIVEGE"
    RELATIVE_LESS = "RELATIVELE"
    RELATIVE_EQUALS = "RELATIVEEE"
    RELATED_GREATER = "MORETHAN"
    RELATED_LESS = "LESSTHAN"


class BooleanOperator(BaseOperator):
    pass


class EmailOperator:
    CHANGES = "VALCHANGES"
    CHANGES_FROM = "CHANGESFROM"
    CHANGES_TO = "CHANGESTO"


class LogicalOperator:
    AND = "^"
    OR = "^OR"
    NQ = "^NQ"



from typing import Any

from . import methods
from .base import BaseRequest


class PatchRequest(BaseRequest):
    _method = methods.PATCH

    def __init__(self, *args: Any, object_id: str, payload: str, **kwargs: Any):
        super(PatchRequest, self).__init__(*args, **kwargs)
        self.payload = payload
        self.object_id = object_id
        self.url_segments.append(object_id)

    def __repr__(self) -> str:
        return self._format_repr(
            f"object_id: {self.object_id}, payload: {self.payload}"
        )

    async def send(self, *args: Any, **kwargs: Any) -> Any:
        return await self._send(data=self.payload, **kwargs)



from typing import Any, Iterable, Union

from aiohttp import ClientResponse, client_exceptions, http_exceptions, web_exceptions
from marshmallow import EXCLUDE

from aiosnow.exceptions import (
    InvalidContentMethod,
    RequestError,
    ServerError,
    UnexpectedResponseContent,
)

from .schemas import ContentSchema


class Response(ClientResponse):
    """Aiosnow Response class

    The Response object holds information about the ServiceNow HTTP response.

    Subclass of aiohttp.ClientResponse, its base reference documentation can be found here:
    https://docs.aiohttp.org/en/latest/client_reference.html#aiohttp.ClientResponse

    Attributes:
        - data: Deserialized (ContentSchema) response content
        - status: HTTP status code of response (int), e.g. 200
        - reason: HTTP status reason of response (str), e.g. "OK"
        - url: Request URL
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super(Response, self).__init__(*args, **kwargs)
        self.data: Union[list, dict, None] = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} {hex(id(self))} {self.url.path} "
            f"[{self.status} {self.reason}]>"
        )

    def __getitem__(self, name: Any) -> Any:
        if isinstance(self.data, dict):
            return self.data.get(name)

        return None

    def __iter__(self) -> Iterable:
        if isinstance(self.data, list):
            yield from self.data
        elif isinstance(self.data, dict):
            yield from self.data.keys()
        else:
            raise InvalidContentMethod(f"Cannot iterate over type: {type(self.data)}")

    def __len__(self) -> int:
        if isinstance(self.data, list):
            return len(self.data)

        return 1

    async def load_document(self) -> None:
        """Deserialize and set response content

        Raises:
            RequestError: If there was an error in the request-response content
        """

        data = await self.json()

        if not isinstance(data, dict):
            if self.status == 204:
                self.data = {}
                return

            await self._handle_error()

        content = ContentSchema(unknown=EXCLUDE, many=False).load(data)

        if "error" in content:
            err = content["error"]
            msg = (
                f"{err['message']}: {err['detail']}"
                if err["detail"]
                else err["message"]
            )

            raise RequestError(msg, self.status)

        self.data = content["result"]

    async def _handle_error(self) -> None:
        """Something went seriously wrong.

        This method interprets the error-response and raises the appropriate exception.

        Raises:
            - ServerError: If the error was interpreted as an unhandled server error
            - UnexpectedResponseContent: If the request was successful, but the request-response contains
            unexpected data
        """

        try:
            # Something went wrong, most likely out of the ServiceNow application's control:
            # Raise exception if we got a HTTP error status back.
            self.raise_for_status()
        except (
            client_exceptions.ClientResponseError,
            http_exceptions.HttpProcessingError,
        ) as exc:
            raise ServerError(exc.message, exc.code) from exc
        except web_exceptions.HTTPException as exc:
            raise ServerError(exc.text or "", exc.status) from exc
        else:
            # Non-JSON content along with a HTTP 200 returned: Unexpected.
            text = await self.text()
            raise UnexpectedResponseContent(
                f"Unexpected response received from server: {text}", 200
            )



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Type

import aiohttp
import marshmallow

from aiosnow.client import Client
from aiosnow.exceptions import InvalidFieldName
from aiosnow.request import (
    DeleteRequest,
    GetRequest,
    PatchRequest,
    PostRequest,
    Response,
    methods,
)

from .schema import ModelSchema, Nested
from .schema.fields import BaseField

req_cls_map = {
    methods.GET: GetRequest,
    methods.POST: PostRequest,
    methods.PATCH: PatchRequest,
    methods.DELETE: DeleteRequest,
}


class BaseModelMeta(type):
    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Any:
        fields = {}

        for base in bases:
            fields.update(base.schema_cls._declared_fields)

        for key, value in attrs.copy().items():
            if isinstance(value, BaseField):
                fields[key] = value
                fields[key].name = key
            elif isinstance(value, marshmallow.schema.SchemaMeta):
                fields[key] = Nested(key, value, allow_none=True, required=False)
            else:
                continue

            # Do not allow override of base members with schema Field attributes.
            for base in bases:
                existing_member = getattr(base, key, None)
                if existing_member is not None and not issubclass(
                    existing_member.__class__,
                    (BaseField, marshmallow.schema.SchemaMeta),
                ):
                    raise InvalidFieldName(
                        f"Field :{name}.{key}: conflicts with a base member, name it something else. "
                        f"The Field :attribute: parameter can be used to give a field an alias."
                    )

        attrs["schema_cls"] = type(name + "Schema", (ModelSchema,), fields)
        cls = super().__new__(mcs, name, bases, attrs)

        return cls


class BaseModel(metaclass=BaseModelMeta):
    """Model base"""

    _session: aiohttp.ClientSession
    _client: Client
    _config: dict = {"return_only": []}
    schema_cls: Type[ModelSchema]
    schema: ModelSchema

    def __init__(self, client: Client):
        self._client = client
        self.fields = dict(self.schema_cls.fields)
        self.nested_fields = {
            n: f for n, f in self.fields.items() if isinstance(f, Nested)
        }
        self.schema = self.schema_cls(unknown=marshmallow.EXCLUDE)
        self._primary_key = getattr(self.schema, "_primary_key")

    @property
    @abstractmethod
    def _api_url(self) -> Any:
        pass

    async def request(self, method: str, *args: Any, **kwargs: Any) -> Response:
        kwargs.update(
            dict(
                api_url=self._api_url,
                session=self._session,
                fields=kwargs.pop("return_only", self._config["return_only"])
                or self.fields.keys(),
            )
        )

        req_cls = req_cls_map[method]
        response = await req_cls(*args, **kwargs).send()

        if method != methods.DELETE:
            response.data = self.schema.load_content(
                response.data, many=isinstance(response.data, list)
            )

        return response

    async def __aenter__(self) -> BaseModel:
        self._session = self._client.get_session()
        return self

    async def __aexit__(self, *_: list) -> None:
        await self._session.close()



from aiosnow.models import TableModel, fields


class IncidentModel(TableModel):
    sys_id = fields.String(is_primary=True)
    impact = fields.IntegerMap()
    number = fields.String()
    parent = fields.String()
    made_sla = fields.Boolean()
    caused_by = fields.String()
    watch_list = fields.String()
    upon_reject = fields.String()
    sys_updated_on = fields.DateTime()
    child_incidents = fields.Integer()
    hold_reason = fields.Integer()
    approval_history = fields.String()
    resolved_by = fields.String()
    sys_updated_by = fields.String()
    opened_by = fields.String()
    user_input = fields.String()
    sys_created_on = fields.DateTime()
    sys_domain = fields.String()
    state = fields.IntegerMap()
    sys_created_by = fields.String()
    knowledge = fields.Boolean()
    order = fields.Integer()
    calendar_stc = fields.Integer()
    closed_at = fields.DateTime()
    cmdb_ci = fields.String()
    delivery_plan = fields.String()
    active = fields.Boolean()
    work_notes_list = fields.String()
    business_service = fields.String()
    priority = fields.IntegerMap()
    sys_domain_path = fields.String()
    rfc = fields.String()
    time_worked = fields.Integer()
    expected_start = fields.DateTime()
    opened_at = fields.DateTime()
    business_duration = fields.DateTime()
    group_list = fields.String()
    work_end = fields.DateTime()
    reopened_time = fields.DateTime()
    resolved_at = fields.DateTime()
    caller_id = fields.String()
    approval_set = fields.DateTime()
    subcategory = fields.String()
    work_notes = fields.String()
    short_description = fields.String()
    close_code = fields.String()
    correlation_display = fields.String()
    work_start = fields.DateTime()
    delivery_task = fields.String()
    assignment_group = fields.String()
    business_stc = fields.Integer()
    additional_assignee_list = fields.String()
    description = fields.String()
    calendar_duration = fields.DateTime()
    notify = fields.Integer()
    sys_class_name = fields.String()
    close_notes = fields.String()
    closed_by = fields.String()
    follow_up = fields.DateTime()
    parent_incident = fields.String()
    contact_type = fields.String()
    reopened_by = fields.String()
    incident_state = fields.IntegerMap()
    urgency = fields.IntegerMap()
    problem_id = fields.String()
    company = fields.String()
    reassignment_count = fields.Integer()
    activity_due = fields.DateTime()
    assigned_to = fields.String()
    severity = fields.IntegerMap()
    comments = fields.String()
    sla_due = fields.DateTime()
    approval = fields.String()
    comments_and_work_notes = fields.String()
    due_date = fields.DateTime()
    sys_mod_count = fields.Integer()
    reopen_count = fields.Integer()
    sys_tags = fields.String()
    escalation = fields.Integer()
    upon_approval = fields.String()
    correlation_id = fields.String()
    location = fields.String()
    category = fields.String()



from typing import Any, Union

from aiosnow.exceptions import AiosnowException, UnexpectedValue
from aiosnow.query.condition import Condition
from aiosnow.query.operators import BaseOperator

from .utils import serialize_list


class BaseQueryable:
    name = None

    def __eq__(self, other: Any) -> Any:
        return self.equals(other)

    def __ne__(self, other: Any) -> Any:
        return self.not_equals(other)

    def __lt__(self, other: Any) -> Condition:
        return self.less_than(other)

    def __le__(self, other: Any) -> Condition:
        return self.less_or_equals(other)

    def __gt__(self, other: Any) -> Condition:
        return self.greater_than(other)

    def __ge__(self, other: Any) -> Condition:
        return self.greater_or_equals(other)

    def equals(self, other: Any) -> Condition:
        raise NotImplementedError

    def not_equals(self, other: Any) -> Condition:
        raise NotImplementedError

    def less_than(self, other: Any) -> Condition:
        raise NotImplementedError

    def less_or_equals(self, other: Any) -> Condition:
        raise NotImplementedError

    def greater_than(self, other: Any) -> Condition:
        raise NotImplementedError

    def greater_or_equals(self, other: Any) -> Condition:
        raise NotImplementedError

    def _condition(
        self, operator: str, value: Union[str, int] = None, field_operator: str = ""
    ) -> Condition:
        if isinstance(value, self.__class__):
            if not field_operator:
                raise UnexpectedValue(
                    f"{operator} does not support field-to-field comparison"
                )

            operator = field_operator
            value = value.name

        if not isinstance(self.name, str):
            raise AiosnowException(f"Missing left operand of {self.__class__}")

        return Condition(self.name, operator, value)

    def in_list(self, values: list) -> Condition:
        """
        Example: impact.in_list([3, 4])

        All records in which the Impact field has one of the values 3 or 4.
        """

        return self._condition(BaseOperator.ONEOF, serialize_list(values))

    def not_in_list(self, values: list) -> Condition:
        """
        Example: impact.not_in_list([3, 4])

        All records in which the Impact field does not have one of the values 3 or 4.
        """

        return self._condition(BaseOperator.NOT_ONEOF, serialize_list(values))

    def is_empty(self) -> Condition:
        """
        All records in which there is no value in the given field
        """

        return self._condition(BaseOperator.EMPTY)

    def is_populated(self) -> Condition:
        """
        All records in which there is any value in the given field
        """

        return self._condition(BaseOperator.POPULATED)

    def is_anything(self) -> Condition:
        """
        All records in which the given field is one of the following:
            - any value
            - empty
            - NULL
        """

        return self._condition(BaseOperator.ANYTHING)



from __future__ import annotations

import json
from typing import Any, AsyncGenerator, Union

from aiosnow.client import Client
from aiosnow.exceptions import (
    DeleteError,
    NoItems,
    PayloadValidationError,
    RequestError,
    SchemaError,
    SelectError,
    TooManyItems,
    UnexpectedResponseContent,
)
from aiosnow.query import Condition, Selector, select
from aiosnow.request import Pagestream, Response, methods

from ..common import BaseModel


class TableModel(BaseModel):
    """Table API model"""

    _config: dict = {"table_name": str, "return_only": []}

    def __init__(self, client: Client, table_name: str, return_only: list = None):
        self._config["table_name"] = table_name
        self._config["return_only"] = return_only or []
        super(TableModel, self).__init__(client)

    @property
    def _api_url(self) -> str:
        return self._client.base_url + "/api/now/table/" + self._config["table_name"]

    async def stream(
        self, selection: Union[Selector, Condition, str] = None, **kwargs: Any
    ) -> AsyncGenerator:
        """Stream-like async generator

        Fetches data in chunks using the ServiceNow pagination system.

        Chunk size determines the number of records to fetch in one go.
        Setting a lower chunk size decreases memory usage, but increases the
        number of requests sent to the server.

        Keyword Args:
            selection: aiosnow-compatible query
            limit (int): Maximum number of records to return
            offset (int): Starting record index
            page_size (int): Number of records to fetch in one go

        Yields:
            Chunk of records
        """

        stream = Pagestream(
            api_url=self._api_url,
            query=select(selection).sysparms,
            session=self._session,
            fields=kwargs.pop("return_only", self._config["return_only"])
            or self.schema.fields.keys(),
            nested_fields=self.nested_fields,
            **kwargs,
        )

        while not stream.exhausted:
            async for response in stream.get_next():
                for record in self.schema.load_content(response.data, many=True):
                    yield response, record

    async def get(
        self, selection: Union[Selector, Condition, str] = None, **kwargs: Any
    ) -> Response:
        """Buffered many

        Fetch and store data in buffer.

        Note: It's recommended to use the stream method when dealing with large
        number of records.

        Keyword Args:
            selection: Aiosnow-compatible query
            limit (int): Maximum number of records to return
            offset (int): Starting record index

        Returns:
            Response
        """

        return await self.request(
            methods.GET,
            query=select(selection).sysparms,
            nested_fields=self.nested_fields,
            **kwargs,
        )

    async def get_one(
        self, selection: Union[Selector, Condition, str], **kwargs: Any
    ) -> Response:
        """Get one record

        Args:
            selection: aiosnow-compatible query

        Returns:
            Response
        """

        if not self._primary_key:
            raise SchemaError(
                f'The targeted "{self.__class__}" cannot be queried: '
                f'its schema lacks a field with "is_primary" set'
            )

        response = await self.get(selection, limit=2, **kwargs)
        if not isinstance(response.data, list):
            raise UnexpectedResponseContent(
                f"Expected a {list} in response to get_one(), got: {type(response.data)}",
                status=response.status,
            )
        elif len(response) > 1:
            raise TooManyItems("Too many results: expected one, got at least 2")
        elif len(response) < 1:
            raise NoItems("Expected a single object in response, got none")

        # Assign the matched record
        response.data = response.data[0]

        return response

    async def get_object_id(self, value: Union[Condition, str]) -> str:
        """Get object id by str or Condition

        Immediately return if value is of str type.

        Args:
            value: Condition or str

        Returns:
            Object id
        """

        if isinstance(value, Condition):
            response = await self.get_one(value, return_only=[self._primary_key])
            return response[self._primary_key]
        elif isinstance(value, str):
            return value
        else:
            raise SelectError(
                f"Selection must be of type {Condition} or {str}, not {type(value)}"
            )

    async def update(self, selection: Union[Condition, str], payload: dict) -> Response:
        """Update matching record

        Args:
            selection: Condition or ID of object to update
            payload: Update payload

        Returns:
            Response
        """

        sys_id = await self.get_object_id(selection)

        if not isinstance(payload, dict):
            raise PayloadValidationError(
                f"Expected payload as a {dict}, got: {type(payload)}"
            )

        return await self.request(
            methods.PATCH, object_id=sys_id, payload=json.dumps(payload),
        )

    async def create(self, payload: dict) -> Response:
        """Create a new record

        Args:
            payload: New record payload

        Returns:
            Response
        """

        return await self.request(methods.POST, payload=json.dumps(payload),)

    async def delete(self, selection: Union[Condition, str]) -> Response:
        """Delete matching record

        Args:
            selection: Condition or ID

        Returns:
            Response
        """

        try:
            sys_id = await self.get_object_id(selection)
        except NoItems:
            raise DeleteError("Cannot delete, no such record")

        response = await self.request(methods.DELETE, object_id=sys_id)

        if response.status != 204:
            text = await response.text()
            raise RequestError(
                f"Unexpected response for DELETE request. "
                f"Status: {response.status}, Text: {text}",
                response.status,
            )

        return response



from marshmallow import Schema, fields


class ErrorSchema(Schema):
    """Defines the structure of the ServiceNow error response content"""

    message = fields.String()
    detail = fields.String(allow_none=True)


class ContentSchema(Schema):
    """Defines structure of the ServiceNow response content"""

    error = fields.Nested(ErrorSchema)
    result = fields.Raw()
    status = fields.String(missing=None)



class AiosnowException(Exception):
    pass


class ConfigurationError(AiosnowException):
    """Configuration error"""


class StreamExhausted(AiosnowException):
    """Signals there are no further items produced by the iterator"""


class NoAuthenticationMethod(AiosnowException):
    """No authentication method was provided"""


class UnexpectedContentType(AiosnowException):
    """Unexpected content type from server"""


class ClientConnectionError(AiosnowException):
    """Raised when there was a problem connecting to the server"""


class SerializationError(AiosnowException):
    """Raised when there was an issue with serialization"""


class DeserializationError(AiosnowException):
    """Raised when there was an issue with deserialization"""


class RequestError(AiosnowException):
    """The application returned an error in the response"""

    def __init__(self, message: str, status: int):
        self.message = message
        self.status = status


class ServerError(RequestError):
    """The server returned an error in the response"""


class UnexpectedResponseContent(RequestError):
    """Unexpected content in response from server"""


class UnexpectedPayloadType(AiosnowException):
    """Raised when the request payload was of an unexpected type"""


class PayloadValidationError(AiosnowException):
    """Local payload validation against a Resource Schema failed"""


class IncompatiblePayloadField(AiosnowException):
    """An incompatible field was found in the payload"""


class UnknownPayloadField(AiosnowException):
    """A field unknown to the schema was found in the payload"""


class IncompatibleSession(AiosnowException):
    """Raised if a custom session object passed to aiosnow.Client is not of aiosnow.Session type"""


class UnexpectedValue(AiosnowException):
    """Typically raised when a method receives unexpected input"""


class SchemaError(AiosnowException):
    """Generic exception raised on schema issues, e.g. integrity errors"""


class SelectError(AiosnowException):
    """Raised on query builder issues"""


class InvalidContentMethod(AiosnowException):
    """Raised if the response content was incorrectly accessed"""


class InvalidFieldName(AiosnowException):
    """Usually raised if an attempt is made to override a base member with a Field"""


class NoSchemaFields(AiosnowException):
    """The schema lacks definition"""


class TooManyItems(AiosnowException):
    """The request yielded too many results"""


class NoItems(AiosnowException):
    """The request yielded no results"""


class DeleteError(AiosnowException):
    """Raised if there was an error deleting a record"""



from typing import Any

from . import methods
from .base import BaseRequest


class DeleteRequest(BaseRequest):
    _method = methods.DELETE

    def __init__(self, *args: Any, object_id: str, **kwargs: Any):
        self.object_id = object_id
        super(DeleteRequest, self).__init__(*args, **kwargs)
        self.url_segments.append(object_id)

    def __repr__(self) -> str:
        return self._format_repr()

    async def send(self, *args: Any, **kwargs: Any) -> Any:
        kwargs["decode"] = False
        return await self._send(*args, **kwargs)



import marshmallow

from aiosnow.query import DateTimeQueryable

from .base import BaseField


class DateTime(marshmallow.fields.DateTime, BaseField, DateTimeQueryable):
    def _bind_to_schema(self, field_name: str, schema: marshmallow.Schema) -> None:
        """Removing this override causes marshmallow schema-field registration to fail
        because self.root resolves to None when inheriting schemas? This
        seems exclusive to the DateTime type, and could be a marshmallow bug.

        "AttributeError: 'NoneType' object has no attribute 'opts'"

        @TODO - find out why DateTime schema parent fails to resolve for DateTime and remove this override"""

        self.format = (
            self.format
            or getattr(schema.opts, self.SCHEMA_OPTS_VAR_NAME)
            or self.DEFAULT_FORMAT
        )



import warnings
from typing import Any, Iterable, Tuple, Union

import marshmallow

from aiosnow.exceptions import (
    DeserializationError,
    IncompatiblePayloadField,
    SchemaError,
    SerializationError,
    UnexpectedPayloadType,
    UnknownPayloadField,
)

from .fields.base import BaseField
from .fields.mapped import MappedField
from .nested import Nested


class ModelSchemaMeta(marshmallow.schema.SchemaMeta):
    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Any:
        attrs["fields"] = fields = {
            k: v for k, v in attrs.items() if isinstance(v, (BaseField, Nested))
        }
        cls = super().__new__(mcs, name, bases, attrs)
        for k, v in fields.items():
            setattr(cls, k, v)

        return cls


class ModelSchema(marshmallow.Schema, metaclass=ModelSchemaMeta):
    @property
    def _primary_key(self) -> Union[str, None]:
        pks = self._pk_candidates

        if len(pks) > 1:
            raise SchemaError(
                f"Multiple primary keys (is_primary) supplied "
                f"in {self.__class__.__name__}. Maximum allowed is 1."
            )
        elif len(pks) == 0:
            return None

        return pks[0]

    @property
    def _pk_candidates(self) -> list:
        return [
            n
            for n, f in self.fields.items()
            if isinstance(f, BaseField) and f.is_primary is True
        ]

    @marshmallow.pre_load
    def _load_response(self, data: Union[list, dict], **_: Any) -> Union[list, dict]:
        """Load response content
        @TODO - move into load()

        Args:
            data: Dictionary of fields to deserialize

        Returns:
            dict(field_name=field_value, ...)
        """

        if isinstance(data, list):
            return [dict(self.__load_response(r or {})) for r in data]
        elif isinstance(data, dict):
            return dict(self.__load_response(data or {}))
        else:
            raise TypeError(
                f"Response content must be {list} or {dict}, got: {type(data)}"
            )

    def __load_response(self, content: dict) -> Iterable[Tuple[str, str]]:
        """Yields deserialized response content items

        Args:
            content: Response content to deserialize

        Yields: <name>, <value>
        """

        for key, value in content.items():
            field = self._declared_fields.get(key, None)

            if not field:
                warnings.warn(
                    f"Unexpected field in response content: {key}, skipping..."
                )
                continue

            if isinstance(field, BaseField):
                if isinstance(value, dict) and {"value", "display_value"} <= set(
                    value.keys()
                ):
                    if isinstance(field, MappedField):
                        value = value["value"] or None, value["display_value"] or None
                    else:
                        value = value[field.pluck.value] or None
            elif isinstance(field, Nested):
                pass
            else:  # Unknown field
                continue

            yield key, value

    def __dump_payload(self, payload: dict) -> Iterable[Tuple[str, str]]:
        """Yields serialized payload

        Args:
            payload: Payload to serialize

        Yields: <name>, <value>
        """

        for key, value in payload.items():
            if isinstance(key, BaseField):
                key = key.name
            elif isinstance(key, str):
                pass
            else:
                raise IncompatiblePayloadField(
                    f"Incompatible field in payload: {type(key)}"
                )

            field = getattr(self, key, None)
            if not field:
                raise UnknownPayloadField(f"Unknown field in payload {key}")

            yield key, value

    def load_content(self, *args: Any, **kwargs: Any) -> dict:
        try:
            return self._do_load(*args, partial=True, postprocess=True, **kwargs)
        except marshmallow.exceptions.ValidationError as e:
            raise DeserializationError(e)

    def loads(self, *args: Any, **kwargs: Any) -> dict:
        try:
            return super().loads(*args, **kwargs)
        except marshmallow.exceptions.ValidationError as e:
            raise DeserializationError(e)

    def dumps(self, obj: dict, *args: Any, many: bool = None, **kwargs: Any) -> str:
        """Dump payload

        Args:
            obj: The object to serialize
            many: Whether to serialize `obj` as a collection. If `None`, the value for `self.many` is used.

        Returns:
            JSON string
        """

        if not isinstance(obj, dict):
            raise UnexpectedPayloadType(
                f"Invalid payload: {type(obj)}, expected: {dict}"
            )

        data = dict(self.__dump_payload(obj))

        try:
            return super().dumps(data)
        except marshmallow.exceptions.ValidationError as e:
            raise SerializationError(e)



from __future__ import annotations

from typing import Any

from aiosnow.consts import SORT_ASCENDING, SORT_DESCENDING


class Selector:
    sysparms: str = ""

    def __init__(self, sysparms: str):
        self.sysparms = sysparms

    def _order_by(self, value: Any, ascending: bool = True) -> str:
        items = value if isinstance(value, list) else [value]
        sort = SORT_ASCENDING if ascending else SORT_DESCENDING
        sort_statements = []
        for value in items:
            if hasattr(value, "name"):
                name = f"{value.name}"
            else:
                name = f"{value}"

            sort_statements.append(sort + name)

        prefix = "^" if self.sysparms else ""
        return prefix + "^".join(sort_statements)

    def order_desc(self, value: Any) -> Selector:
        self.sysparms += self._order_by(value, ascending=False)
        return self

    def order_asc(self, value: Any) -> Selector:
        self.sysparms += self._order_by(value)
        return self

    def __repr__(self) -> Any:
        return self.sysparms



from .condition import Condition
from .fields import (
    BooleanQueryable,
    DateTimeQueryable,
    IntegerQueryable,
    StringQueryable,
)
from .operators import (
    BaseOperator,
    BooleanOperator,
    DateTimeOperator,
    EmailOperator,
    IntegerOperator,
    LogicalOperator,
    ReferenceField,
    StringOperator,
)
from .selector import Selector
from .utils import select



from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from typing import Any, List, Tuple
from urllib.parse import urlencode, urlparse

from aiohttp import client_exceptions

from aiosnow.consts import CONTENT_TYPE
from aiosnow.exceptions import ClientConnectionError, UnexpectedContentType
from aiosnow.request import methods
from aiosnow.session import Session

from .response import Response


class BaseRequest(ABC):
    session: Session
    log = logging.getLogger("aiosnow.request")

    def __init__(
        self, api_url: str, session: Session, fields: list = None,
    ):
        self.api_url = api_url
        self.session = session
        self.fields = [str(f) for f in fields or []]
        self.url_segments: List[str] = []
        self.headers_default = {"Content-type": CONTENT_TYPE}
        self._req_id = f"REQ_{hex(int(round(time.time() * 1000)))}"

    @property
    def url_params(self) -> dict:
        params = dict(sysparm_display_value="all")
        if self.fields:
            params["sysparm_fields"] = ",".join(self.fields)

        return params

    @property
    def url(self) -> str:
        api_url = self.api_url

        if self.url_segments:
            # Append path segments
            api_url += "/" + "/".join(map(str, self.url_segments))

        return f"{api_url}?{urlencode(self.url_params)}"

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    async def send(self, *args: Any, **kwargs: Any) -> Tuple[Response, dict]:
        pass

    @property
    @abstractmethod
    def _method(self) -> str:
        pass

    @property
    def _request_id(self) -> str:
        return hex(id(self))

    def _format_repr(self, params: str = "") -> str:
        return f"<{self.__class__.__name__} {urlparse(self.url).path} [{params}]>"

    async def _send(
        self, headers_extra: dict = None, decode: bool = True, **kwargs: Any,
    ) -> Response:
        headers = self.headers_default
        headers.update(**headers_extra or {})
        kwargs["headers"] = headers

        method = kwargs.pop("method", self._method)

        try:
            self.log.debug(f"{self._req_id}: {self}")
            response = await self.session.request(method, self.url, **kwargs)
            self.log.debug(f"{self._req_id}: {response}")
        except client_exceptions.ClientConnectionError as exc:
            raise ClientConnectionError(str(exc)) from exc

        if method == methods.DELETE and response.status == 204:
            return response

        if not response.content_type.startswith(CONTENT_TYPE):
            raise UnexpectedContentType(
                f"Unexpected content-type in response: "
                f"{response.content_type}, expected: {CONTENT_TYPE}, "
                f"probable causes: instance down or REST API disabled"
            )

        if decode:
            await response.load_document()
        else:
            response.data = await response.read()

        return response



from __future__ import annotations

from typing import Union

from .operators import LogicalOperator


class Condition:
    def __init__(self, key: str, operator: str, value: Union[str, int, None]):
        self.operand_left = key
        self.operand_right = value
        self.operator_conditional = operator
        self.operator_logical = ""
        self.registry = [self]

    def __str__(self) -> str:
        return self.serialize()

    def serialize(self, cond: Condition = None) -> str:
        c = cond or self
        return "".join(
            [
                c.operator_logical,
                c.operand_left,
                c.operator_conditional,
                str(c.operand_right),
            ]
        )

    def serialize_registry(self) -> str:
        """Condition string representation

        Returns: sysparm query
        """

        chain = ""

        for c in self.registry:
            chain += self.serialize(c)

        return chain

    def _merge_registry(self, registry: list) -> None:
        self.registry += registry

    def _add_condition(self, cond: Condition, operator: str) -> Condition:
        """Adds a new condition to chain

        Args:
            cond: Condition
            operator: Logical operator

        Returns: self
        """

        cond.operator_logical = str(operator)
        self._merge_registry(cond.registry)

        return self

    def __and__(self, next_cond: Condition) -> Condition:
        """Appends ^ Condition to chain"""

        return self._add_condition(next_cond, LogicalOperator.AND)

    def __or__(self, next_cond: Condition) -> Condition:
        """Appends ^OR Condition to chain"""

        return self._add_condition(next_cond, LogicalOperator.OR)

    def __xor__(self, next_cond: Condition) -> Condition:
        """Appends ^NQ Condition to chain"""

        return self._add_condition(next_cond, LogicalOperator.NQ)



from abc import ABC
from typing import Any, Tuple, Union

import marshmallow

from .integer import Integer
from .string import String


class Mapping(ABC):
    key: Union[str, int, None] = None
    value: Union[str, None] = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} [key={self.key}, value={self.value}]>"


class StringMapping(Mapping):
    def __init__(self, key: str, value: str):
        self.key = key
        self.value = value


class IntegerMapping(Mapping):
    def __init__(self, key: int, value: str):
        self.key = key
        self.value = value


class MappedField(marshmallow.fields.Tuple):
    def __init__(self, *args: Any, dump_text: bool = False, **kwargs: Any):
        self.should_dump_text = dump_text
        super(MappedField, self).__init__(self._tuple_fields, *args, **kwargs)

    @property
    def _tuple_fields(self) -> Tuple[Union[Integer, String], String]:
        raise NotImplementedError

    def _serialize(self, *args: Any, **kwargs: Any) -> Any:
        obj = args[0]

        if self.should_dump_text:
            return obj.value

        return obj.key


class IntegerMap(MappedField, Integer):
    def _deserialize(self, *args: Any, **kwargs: Any) -> Any:
        result = super()._deserialize(*args, **kwargs)
        return IntegerMapping(*result)

    def _serialize(self, *args: Any, **kwargs: Any) -> Any:
        return super()._serialize(*args, **kwargs)

    _tuple_fields = Integer(), String()


class StringMap(MappedField, String):
    def _deserialize(self, *args: Any, **kwargs: Any) -> Any:
        result = super()._deserialize(*args, **kwargs)
        return StringMapping(*result)

    def _serialize(self, *args: Any, **kwargs: Any) -> Any:
        return super()._serialize(*args, **kwargs)

    _tuple_fields = String(), String()



from typing import Any

import marshmallow

from aiosnow.models.common.schema.helpers import Pluck


class BaseField(marshmallow.fields.Field):
    def __init__(
        self,
        *args: Any,
        pluck: Pluck = Pluck.VALUE,
        is_primary: bool = False,
        **kwargs: Any,
    ):
        self.pluck = Pluck(pluck)
        self.is_primary = is_primary
        super(BaseField, self).__init__(*args, **kwargs)
        self.allow_none = True
        self.missing = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} [maps_to={self.pluck}, primary={self.is_primary}]>"



from typing import Any

from . import methods
from .base import BaseRequest


class PostRequest(BaseRequest):
    _method = methods.POST

    def __init__(self, *args: Any, payload: str, **kwargs: Any):
        self.payload = payload
        super(PostRequest, self).__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return self._format_repr(f"payload: {self.payload}")

    async def send(self, *args: Any, **kwargs: Any) -> Any:
        return await self._send(data=self.payload, **kwargs)



from typing import Any

import aiohttp

from aiosnow.request.response import Response


class Session(aiohttp.ClientSession):
    def __init__(self, *args: Any, **kwargs: Any):
        super(Session, self).__init__(
            *args, response_class=kwargs.pop("response_class", Response), **kwargs
        )

    async def request(self, *args, **kwargs) -> Any:  # type: ignore
        return await super().request(*args, **kwargs)



from typing import Any, AsyncGenerator

from aiosnow.consts import DEFAULT_LIMIT, DEFAULT_PAGE_SIZE
from aiosnow.request import GetRequest


class Pagestream(GetRequest):
    """Aiosnow Pagestream

    The Pagestream object behaves much like a stream, it's memory friendly and yields deserialized
    chunks of records using the ServiceNow pagination system.

    Attributes:
        page_size (int): Number of records in each page
    """

    exhausted = False

    def __init__(self, *args: Any, page_size: int = DEFAULT_PAGE_SIZE, **kwargs: Any):
        self._limit_total = kwargs.pop("limit", DEFAULT_LIMIT)
        self._offset_initial = kwargs.get("offset", 0)
        super(Pagestream, self).__init__(*args, **kwargs)
        self._limit = page_size

    def __repr__(self) -> str:
        params = (
            f"query: {self.query or None}, limit: {self._limit_total}, "
            f"offset: {self.offset}, page_size: {self._limit}"
        )
        return self._format_repr(params)

    async def get_next(self, **kwargs: Any) -> AsyncGenerator:
        if self.exhausted:
            return

        offset_limit = self._limit_total + self._offset_initial

        # Max number of records is less than or equals the
        # requested page size: set exhausted adjust the page size
        if offset_limit <= self.limit:
            self._limit = self._limit_total
            self.exhausted = True
        # Offset is greater or equals the offset limit: set exhausted
        # and adjust the final page size
        elif self.offset >= offset_limit:
            self.exhausted = True
            if self.offset == offset_limit:  # No need to fetch an empty page
                return

            # Adjust limit to get the last X records
            # This could be a fraction of a really large page
            self._limit = offset_limit - self._limit_total
            self.log.debug(
                f"{self._req_id}: Limit of {self.limit} records reached, setting exhausted"
            )

        response = await self.send(**kwargs)
        self._offset += self._limit

        if "next" not in response.links:
            self.exhausted = True
            self.log.debug(f"{self._req_id}: No more pages, setting exhausted")

        await response.load_document()
        yield response



from aiosnow.query.condition import Condition
from aiosnow.query.operators import DateTimeOperator

from .base import BaseQueryable


class DateTimeQueryable(BaseQueryable):
    def on(self, value: str) -> Condition:
        """
        Example: sla_due.on("2019-12-24 02:03:04")

        All records in which the value for the SLA due field matches the given date
        """

        return self._condition(DateTimeOperator.ON, value)

    def not_on(self, value: str) -> Condition:
        """
        Example: sla_due.not_on("2019-12-24 02:03:04")

        All records in which the value for the SLA due field is any other but the given date
        """

        return self._condition(DateTimeOperator.NOT_ON, value)

    def before(self, value: str) -> Condition:
        """
        Example: sla_due.before("2019-12-24 02:03:04")

        All records in which the value for the SLA due field is any date previous to the given date.
        """

        return self._condition(DateTimeOperator.LESS, value)

    def after(self, value: str) -> Condition:
        """
        Example: sla_due.after("2019-12-24 02:03:04")

        All records in which the value for the SLA due field is any date after the given date.
        """

        return self._condition(DateTimeOperator.GREATER, value)

    def until(self, value: str) -> Condition:
        """
        Example: sla_due.until("2019-12-24 02:03:04")

        All records in which the value for the SLA due field is one of the following:
            - any date previous to today
            - today
        """

        return self._condition(DateTimeOperator.LESS_EQUALS, value)

    def as_of(self, value: str) -> Condition:
        """
        Example: sla_due.as_of("2019-12-24 02:03:04")

        All records in which the date value for the SLA due field is one of the following:
            - today
            - any date after today
        """

        return self._condition(DateTimeOperator.GREATER_EQUALS, value)

    def between(self, value1: str, value2: str) -> Condition:
        """
        Example:

        All records in which the value for the SLA due field is between the given dates
        """

        return self._condition(DateTimeOperator.BETWEEN, f"{value1}@{value2}")



from aiosnow.query.condition import Condition
from aiosnow.query.operators import StringOperator

from .base import BaseQueryable


class StringQueryable(BaseQueryable):
    def equals(self, value: str) -> Condition:
        """
        Example: short_description.equals("Network storage is unavailable")

        All records in which the Short description says nothing else but "Network storage is unavailable."
        """

        return self._condition(
            StringOperator.EQUALS, value, field_operator=StringOperator.SAME
        )

    def not_equals(self, value: str) -> Condition:
        """
        Example: short_description.not_equals("Network storage is unavailable")

        All records in which the value for the Short description field says anything but "Network storage
        is unavailable."
        """

        return self._condition(
            StringOperator.NOT_EQUALS, value, field_operator=StringOperator.DIFFERENT
        )

    def starts_with(self, value: str) -> Condition:
        """
        Example: short_description.starts_with("SAP")

        All records in which the characters "SAP" appear at the beginning of the value for the Short description field.
        """

        return self._condition(StringOperator.STARTSWITH, value)

    def ends_with(self, value: str) -> Condition:
        """
        Example: short_description.ends_with("outage")

        All records in which the string "outage" appears at the end of the value for the Short description field.
        """

        return self._condition(StringOperator.ENDSWITH, value)

    def contains(self, value: str) -> Condition:
        """
        Example: short_description.contains("SAP")

        All records in which the characters "SAP" appear anywhere in the value for the Short description field.
        """

        return self._condition(StringOperator.CONTAINS, value)

    def not_contains(self, value: str) -> Condition:
        """
        Example: short_description.not_contains("SAP")

        All records in which the characters "SAP" do not appear anywhere in the value for the Short description field.
        """

        return self._condition(StringOperator.NOT_CONTAINS, value)

    def less_or_equals(self, value: str) -> Condition:
        """
        Example: short_description.less_or_equals("s")

        All records in which the string in the Short description field is one of the following:
            - the first letter is any letter between "a" and "s"
            - the exact value is "s"
        """

        return self._condition(StringOperator.LESS_EQUALS, value)

    def greater_or_equals(self, value: str) -> Condition:
        """
        Example: short_description.greater_or_equals("s")

        All records in which the string in the Short description field is one of the following:
            - the first letter is any letter between "s" and "z"
            - the exact value is "s"
        """

        return self._condition(StringOperator.GREATER_EQUALS, value)

    def between(self, value1: str, value2: str) -> Condition:
        """
        Example: short_description.between("q", "t")

        All records in which the first letter in the Short description field is "q," "r," "s," or "t."
        """

        value = f"{value1}@{value2}"
        return self._condition(StringOperator.BETWEEN, value)



from aiosnow.query.condition import Condition
from aiosnow.query.operators import BooleanOperator

from .base import BaseQueryable


class BooleanQueryable(BaseQueryable):
    def is_true(self) -> Condition:
        """
        Example: active.is_true()

        All records in which the Active field is True.
        """

        return self._condition(BooleanOperator.EQUALS, "true")

    def is_falsy(self) -> Condition:
        """
        Example: active.is_falsy()

        All records in which the Active field is False, empty, or NULL.
        """

        return self._condition(BooleanOperator.NOT_EQUALS, "true")



from __future__ import annotations

import json
from typing import Any

from marshmallow import Schema, ValidationError, fields, post_load

from aiosnow.exceptions import ConfigurationError


class ConfigEncoder(json.JSONEncoder):
    def default(self, o: BaseConfigSchema.ConfigSegment) -> dict:
        return o.as_dict()


class BaseConfigSchema(Schema):
    class ConfigSegment:
        """Internal config segment"""

        def __repr__(self) -> str:
            return json.dumps(self.__config, cls=ConfigEncoder, indent=4)

        def as_dict(self) -> dict:
            return self.__config

        def __init__(self, **config: dict):
            self.__config = config
            for k, v in config.items():
                setattr(self, k, v)

    @post_load
    def make_object(self, data: dict, **_: Any) -> Any:
        try:
            return self.ConfigSegment(**data)
        except ValidationError as exc:
            raise ConfigurationError from exc


class SessionConfig(BaseConfigSchema):
    """Session config schema

    Attributes:
        address (str): Instance address, e.g. my_instance.service-now.com
        basic_auth (tuple): (<username>, <password>), mutually exclusive with other authentication methods
        use_ssl (bool): Whether to use SSL, defaults to True
        verify_ssl (bool): Whether to validate SSL certificates, defaults to True
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super(SessionConfig, self).__init__(*args, **kwargs)

    basic_auth = fields.Tuple(
        tuple_fields=(fields.String(), fields.String()), required=False, allow_none=True
    )
    use_ssl = fields.Boolean(missing=True)
    verify_ssl = fields.Boolean(missing=True)

    @post_load
    def make_object(self, data: dict, **_: Any) -> Any:
        if {"basic_auth", "oauth"} <= set(data):
            raise ValidationError("Cannot use multiple authentication methods")
        elif data.get("basic_auth"):
            pass
        else:
            raise ValidationError("No supported authentication method provided")

        return super().make_object(data)


class ConfigSchema(BaseConfigSchema):
    """Aiosnow config schema

    Attributes:
        session (bool): Session config
    """

    address = fields.String(required=True)
    session = fields.Nested(
        SessionConfig, default={}, allow_none=True
    )  # type: SessionConfig # type: ignore

    def __init__(self, *args: Any, **kwargs: Any):
        super(ConfigSchema, self).__init__(*args, **kwargs)

    def load(self, *args: Any, **kwargs: Any) -> ConfigSchema:
        return super().load(*args, **kwargs)



from __future__ import annotations

from typing import Any, Dict, Iterable, Union
from urllib.parse import urlparse

from . import methods
from .base import BaseRequest

_cache: dict = {}


class GetRequest(BaseRequest):
    _method = methods.GET

    def __init__(
        self,
        *args: Any,
        nested_fields: dict = None,
        limit: int = 10000,
        offset: int = 0,
        query: str = None,
        **kwargs: Any,
    ):
        self.nested_fields = nested_fields or {}
        self.nested_attrs = list(self._nested_attrs)
        self._limit = offset + limit
        self._offset = offset
        self.query = query
        super(GetRequest, self).__init__(*args, **kwargs)

    def __repr__(self) -> str:
        params = (
            f"query: {self.query or None}, limit: {self.limit}, offset: {self.offset}"
        )
        return self._format_repr(params)

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def limit(self) -> int:
        return self._limit

    @property
    def _nested_attrs(self) -> Iterable:
        for field in self.nested_fields.values():
            yield from field.nested.fields.keys()

    async def _expand_nested(
        self, content: Union[dict, list, None]
    ) -> Union[dict, list, None]:
        if not self.nested_fields:
            pass
        elif isinstance(content, dict):
            nested = await self._resolve_nested(content)
            content.update(nested)
        elif isinstance(content, list):
            for idx, record in enumerate(content):
                nested = await self._resolve_nested(record)
                content[idx].update(nested)

        return content

    async def _resolve_nested(self, content: dict) -> dict:
        nested: Dict[Any, Any] = {}
        nested_attrs = list(self.nested_attrs)

        for field_name in self.nested_fields.keys():
            item = content[field_name]
            if not item or not item["display_value"]:
                continue
            elif "link" not in item:
                nested[field_name] = item
                continue

            nested[field_name] = await self.get_cached(item["link"], nested_attrs)

        return nested

    async def get_cached(self, url: str, fields: list) -> dict:
        if url not in _cache:
            record_id = urlparse(url).path.split("/")[-1]
            request = GetRequest(url, self.session, fields=fields)
            response = await request._send(method=methods.GET)
            self.log.debug(f"Caching response for: {record_id}")
            _cache[url] = response.data

        return _cache[url]

    async def send(self, *args: Any, resolve: bool = True, **kwargs: Any) -> Any:
        response = await self._send(**kwargs)
        if resolve:
            response.data = await self._expand_nested(response.data)

        return response

    @property
    def url_params(self) -> dict:
        return dict(
            sysparm_offset=self.offset,
            sysparm_limit=self.limit,
            sysparm_query=self.query,
            **super().url_params,
        )



from typing import Union

from aiosnow.exceptions import SelectError

from .condition import Condition
from .selector import Selector


def select(value: Union[Selector, Condition, str] = None) -> Selector:
    if value is None or isinstance(value, str):
        sysparms = value or ""
    elif isinstance(value, Condition):
        sysparms = value.serialize_registry()
    elif isinstance(value, Selector):
        return value
    else:
        raise SelectError(f"Can only query by type {Condition} or {str}")

    return Selector(sysparms)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosnow',
 'aiosnow.models',
 'aiosnow.models.common',
 'aiosnow.models.common.schema',
 'aiosnow.models.common.schema.fields',
 'aiosnow.models.common.schema.helpers',
 'aiosnow.models.table',
 'aiosnow.models.table.declared',
 'aiosnow.query',
 'aiosnow.query.fields',
 'aiosnow.request',
 'aiosnow.request.helpers',
 'aiosnow.request.response']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0', 'marshmallow>=3.6.1,<4.0.0']

setup_kwargs = {
    'name': 'aiosnow',
    'version': '0.5.2',
    'description': 'Asynchronous Python ServiceNow library',
    'long_description': '# aiosnow: Asynchronous Python ServiceNow Library\n\n[![image](https://badgen.net/pypi/v/aiosnow)](https://pypi.org/project/aiosnow)\n[![image](https://badgen.net/badge/python/3.7+?color=purple)](https://pypi.org/project/aiosnow)\n[![image](https://badgen.net/travis/rbw/aiosnow)](https://travis-ci.org/rbw/aiosnow)\n[![image](https://badgen.net/pypi/license/aiosnow)](https://raw.githubusercontent.com/rbw/aiosnow/master/LICENSE)\n[![image](https://pepy.tech/badge/snow/month)](https://pepy.tech/project/snow)\n\n**aiosnow** is a Python [asyncio](https://docs.python.org/3/library/asyncio.html) library for interacting with ServiceNow programmatically. It hopes to be:\n\n- Convenient: A good deal of work is put into making **aiosnow** flexible and easy to use.\n- Performant: Remote API calls uses non-blocking sockets tracked by an event loop, allowing large amounts of lightweight request tasks to run concurrently.\n- Modular: Core functionality is componentized into modules that are built with composability and extensibility in mind.\n\n*Example code*\n\n```python\nimport asyncio\n\nimport aiosnow\nfrom aiosnow.models.table.declared import IncidentModel as Incident\n\nasync def main():\n    client = aiosnow.Client("<instance>.service-now.com", basic_auth=("<username>", "<password>"))\n\n    async with Incident(client, table_name="incident") as inc:\n        # Fetch high-priority incidents\n        for response in await inc.get(Incident.priority <= 3, limit=5):\n            print(f"Number: {response[\'number\']}, Priority: {response[\'priority\'].value}")\n\nasyncio.run(main())\n```\n\nCheck out the [examples directory](examples) for more material.\n\n### Documentation\n\nAPI reference and more is available in the [technical documentation](https://aiosnow.readthedocs.io/en/latest).\n\n\n### Funding\n\nThe **aiosnow** code is permissively licensed, and can be incorporated into any type of application–commercial or otherwise–without costs or limitations.\nIts author believes it\'s in the commercial best-interest for users of the project to invest in its ongoing development.\n\nConsider leaving a [donation](https://paypal.vault13.org) if you like this software, it will:\n\n- Directly contribute to faster releases, more features, and higher quality software.\n- Allow more time to be invested in documentation, issue triage, and community support.\n- Safeguard the future development of **aiosnow**.\n\n### Development status\n\nBeta: Core functionality is done and API breakage unlikely to happen.\n\n\n### Contributing\n\nCheck out the [contributing guidelines](CONTRIBUTING.md) if you want to help out with code or documentation.\n\n\n### Author\n\nRobert Wikman \\<rbw@vault13.org\\>\n\n',
    'author': 'Robert Wikman',
    'author_email': 'rbw@vault13.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rbw/aiosnow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from aiosnow.models import TableModel, fields


class JournalModel(TableModel):
    sys_id = fields.String(is_primary=True)
    sys_created_on = fields.DateTime()
    sys_created_by = fields.String()
    element = fields.String()
    element_id = fields.String()
    name = fields.String()
    value = fields.String()
    sys_tags = fields.String()



from typing import Dict, Union

import aiohttp
from marshmallow import ValidationError

from aiosnow.config import ConfigSchema
from aiosnow.exceptions import (
    ConfigurationError,
    IncompatibleSession,
    NoAuthenticationMethod,
)
from aiosnow.request import Response
from aiosnow.session import Session
from aiosnow.utils import get_url


class Client:
    """aiosnow client

    Args:
        address: Instance TCP address, example: my-instance.service-now.com
        basic_auth: Tuple of (username, password)
        use_ssl: Whether to use SSL
        verify_ssl: Whether to verify SSL certificates
        session: Custom aiohttp.ClientSession object

    Attributes:
        config: Client configuration object
    """

    _preconf_session = None

    def __init__(
        self,
        address: Union[str, bytes],
        basic_auth: tuple = None,
        use_ssl: bool = True,
        verify_ssl: bool = None,
        session: aiohttp.ClientSession = None,
    ):
        app_config: Dict = dict(address=address)

        if session:
            if not isinstance(session, aiohttp.ClientSession):
                raise IncompatibleSession(
                    f"The aiosnow.Client expects :session: to be a aiosnow-compatible "
                    f"{aiohttp.ClientSession}, not {session}"
                )

            resp_cls = getattr(session, "_response_class")
            if resp_cls != Response:
                raise IncompatibleSession(
                    f"The {session} passed to {self} must have its :response_class: "
                    f"set to {Response}, not {resp_cls}"
                )

            session_config_params = [basic_auth, verify_ssl]
            if any([p is not None for p in session_config_params]):
                raise ConfigurationError(
                    f"Client Session factory configuration params {session_config_params} "
                    f"cannot be used with a custom Session object."
                )

            self._preconf_session = session
        else:
            app_config["session"] = dict(
                basic_auth=basic_auth,
                use_ssl=use_ssl or True,
                verify_ssl=verify_ssl or True,
            )

        try:
            self.config = ConfigSchema(many=False).load(app_config)
        except ValidationError as e:
            raise ConfigurationError(e)

        self.base_url = get_url(str(self.config.address), bool(use_ssl))

    @property
    def _auth(self) -> aiohttp.BasicAuth:
        """Get authentication object built using config

        Returns:
            aiohttp-compatible authentication object
        """

        if self.config.session.basic_auth:
            return aiohttp.BasicAuth(*self.config.session.basic_auth)  # type: ignore
        else:
            raise NoAuthenticationMethod("No known authentication methods was provided")

    def get_session(self) -> aiohttp.ClientSession:
        """New client session

        Returns:
            aiohttp.ClientSession:  HTTP client session
        """

        if self._preconf_session:
            return self._preconf_session

        connector_args: dict = {}

        if self.config.session.use_ssl:
            connector_args["verify_ssl"] = self.config.session.verify_ssl

        return Session(
            auth=self._auth, connector=aiohttp.TCPConnector(**connector_args)
        )



from typing import Any

import marshmallow


class Nested(marshmallow.fields.Nested):
    def __init__(self, parent_name: str, nested_cls: type, *args: Any, **kwargs: Any):
        fields = getattr(nested_cls, "_declared_fields", {})

        for name, field in fields.items():
            field.name = f"{parent_name}.{name}"
            setattr(self, name, field)

        super(Nested, self).__init__(nested_cls, *args, **kwargs)
